/*
    Disk throughput benchmarking tool
    Copyright (C) 2006 Darrick Wong

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "bogodisk.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <stdint.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mount.h>

static FILE *tty;

static void print_help(void) {
	printf("Usage: %s [options] device [devices...] [-o report]\n", PACKAGE);
	printf("\n");
	printf("Options\n");
	printf(" -b	Start test at this location.\n");
	printf(" -e	End test after this location.\n");
	printf(" -c	Use memory buffer of this size.\n");
	printf(" -o	Write output to the file \"report\".\n");
}

static uint64_t get_number(const char *string)
{
	int sz = strlen(string);
	double result = atof(string);
	char last = string[sz - 1];

	switch (last) {
		case 'k':
		case 'K':
			result *= 1024;
			break;
		case 'm':
		case 'M':
			result *= 1048576;
			break;
		case 'g':
		case 'G':
			result *= 1073741824;
			break;
	}

	return result;
}

static uint32_t get_block_size(const char *fname) {
	struct stat devinfo;
	uint32_t blksz;

	if (!stat(fname, &devinfo))
		return devinfo.st_blksize;

	perror(fname);
	fprintf(stderr, "%s: Unknown block size; guessing page size alignment?\n",
		fname);

#ifdef _SC_PAGESIZE
	blksz = sysconf(_SC_PAGESIZE);
	if (blksz > 0)
		goto found_blksz;
#endif
#ifdef PAGESIZE
	blksz = sysconf(PAGESIZE);
	if (blksz > 0)
		goto found_blksz;
#endif
	blksz = 4096;
found_blksz:

	return blksz;
}

static uint64_t get_size(const char *fname)
{
	uint64_t res = 0;
	int is_block = 0;
	int fd;
#ifdef __USE_LARGEFILE64
	struct stat64 si64;
#else
	struct stat si;
#endif

#ifdef __USE_LARGEFILE64
	if (!stat64(fname, &si64)) {
		res = si64.st_size;
		if (S_ISBLK(si64.st_mode))
			is_block = 1;
	}
#else
	if (stat(fname, &si)) {
		res = si.st_size;
		if (S_ISBLK(si.st_mode))
			is_block = 1;
	}
#endif

	if (is_block) {
#ifdef BLKGETSIZE64
		fd = open(fname, O_RDONLY);
		if (fd < 0) {
			perror(fname);
			return 0;
		}
		if (ioctl(fd, BLKGETSIZE64, &res)) {
			perror(fname);
			return 0;
		}
		close(fd);
#else
		if (!res)
			fprintf("Don't know how to get size of block "
				"device!\n");
#endif
	}
	
	return res;
}

static void time_device(const char *dev, uint64_t start, uint64_t stop,
			uint64_t bufsize, FILE *report)
{
	FILE *out = stdout;
	uint64_t max_size, pos;
	void *buf;
	struct timespec now, before, rep;
	double nowf, beforef, repf;
	int clock_type = CLOCK_MONOTONIC;
	int fd, ret, blksz;

	/* Flip backwards params */
	if (start && stop && (start > stop)) {
		max_size = start;
		start = stop;
		stop = max_size;
	}

	/* Check parameters */
	max_size = get_size(dev);
	if (!max_size) {
		fprintf(stderr, "%s: Size is zero, skipping.\n", dev);
		return;
	}

	if (start > max_size) {
		fprintf(stderr, "%s: Starting at %llu, which is beyond the "
			"end of the device at %llu!\n", dev, start, max_size);
		return;
	}

	fprintf(out, "%s: Max size %llu\n", dev, max_size);

	if (stop > max_size) {
		fprintf(stderr, "%s: Clamping end to %llu.\n", dev, max_size);
		stop = max_size;
	} else if (!stop) {
		stop = max_size;
	}
	
	if (report == stdout)
		out = tty;
	else
		fprintf(out, "%s: Reading from %llu to %llu in %llu byte "
			"chunks.\n", dev, start, stop, bufsize);
	fprintf(report, "%s: Reading from %llu to %llu in %llu byte "
		"chunks.\n", dev, start, stop, bufsize);

	blksz = get_block_size(dev);
	fprintf(out, "%s: Block size %d bytes.\n", dev, blksz);

	/* grab buffer */
	if (posix_memalign(&buf, blksz, bufsize)) {
		perror("posix_memalign");
		return;
	}

	/* lock pages */
#ifdef _POSIX_MEMLOCK_RANGE
	if (mlock(buf, bufsize)) {
		perror("mlock");
		fprintf(stderr, "Can't lock pages; proceeding anyway.\n");
	}
#endif

	/* open device */
	fd = open(dev, O_RDONLY | O_DIRECT | O_LARGEFILE);
	if (!fd) {
		perror("dev");
		goto unlock_mem;
	}

	/* make sure CLOCK_MONOTONIC is supported */
	if (clock_gettime(clock_type, &now)) {
		clock_type = CLOCK_REALTIME;
	}

	/* seek to starting point */
	if (start) {
		pos = lseek64(fd, start, SEEK_SET);
		if (pos != start) {
			perror("lseek");
			goto end_test;
		}
	} else {
		pos = 0;
	}

	clock_gettime(clock_type, &rep);
	repf = (rep.tv_sec) + ((double)rep.tv_nsec / 1000000000);

	/* loop */
	while (pos < stop) {
		clock_gettime(clock_type, &before);
		ret = read(fd, buf, bufsize);
		if (!ret)
			break;
		else if (ret < 0) {
			perror(dev);
			if (ret != EIO)
				break;
			ret = 512;
		}
		clock_gettime(clock_type, &now);
		nowf = (now.tv_sec) + ((double)now.tv_nsec / 1000000000);
		beforef = (before.tv_sec) + ((double)before.tv_nsec / 1000000000);
		fprintf(report, "%llu, %llu, %.2f, %.2f\n", pos, pos + ret,
			(nowf - beforef),
			(double)ret / (nowf - beforef));
		if ((out == stdout) && ((nowf - repf) > 5)) {
			fprintf(out, "%s: Passed %llu (%.2f%%)            \r",
				dev, pos + ret, (double)100 *
				(pos + ret - start) / (stop - start));
			fflush(out);
			repf = nowf;
		}
		pos += ret;
	}

	if (out == stdout)
		fprintf(out, "\n");

end_test:
	/* close device */
	close(fd);

unlock_mem:
	/* unlock pages */
#ifdef _POSIX_MEMLOCK_RANGE
	if (munlock(buf, bufsize)) {
		perror("munlock");
	}
#endif
	free(buf);
}

int main(int argc, char *argv[])
{
	uint64_t start, stop, bufsize;
	int i;
	FILE *report = stdout;
	char c;

	start = stop = 0;
	bufsize = DEFAULT_BUFSIZE;

	tty = fopen("/dev/tty", "a+");
	if (!tty) {
		perror("/dev/tty");
		return 1;
	}
	fprintf(tty, "%s %s, Copyright (C) 2006 Darrick Wong.\n",
		PACKAGE, PACKAGE_VERSION);

	/* parse args */
	while((c = getopt(argc, argv, "b:e:c:o:")) != -1) {
		switch(c) {
			case 'o':
				report = fopen(optarg, "w+");
				if (!report) {
					perror(optarg);
					return 2;
				}
				break;
			case 'b':
				start = get_number(optarg);
				break;
			case 'e':
				stop = get_number(optarg);
				break;
			case 'c':
				bufsize = get_number(optarg);
				break;
			default:
				print_help();
				return 1;
		}
	}

	if (argc == optind)
		print_help();

	for (i = optind; i < argc; i++) {
		time_device(argv[i], start, stop, bufsize, report);
	}

	return 0;
}
